

/*
	org				: 115.6
	+ ints & temps	: 96.9	(20%)
	+ cnd-swap		: 82.0	(18%)
	+ inline stack	: 76.4	( 7%)
	+ clrUp/Down	: 69.4	(10%)
	Total: 67% faster
	---
	floodfill_asm	: 67.0
	/ DMA			: 68.4
	/ strh during search	: 69.6
	
*/


// Includes
#include <PA9.h> // Include for PA_Lib
#include "registers_alt.h"

#define INLINE static inline

//! Start a profiling run
/*!	\note Routine uses timers 3 and 3; if you're already using these
*	  somewhere, chaos is going to ensue.
*/
INLINE void profile_start()
{
	REG_TM2D= 0;	REG_TM3D= 0;
    REG_TM2CNT= 0;	REG_TM3CNT= 0;
    REG_TM3CNT= TIMER_ENABLE | TIMER_CASCADE;
    REG_TM2CNT= TIMER_ENABLE;
}

//! Stop a profiling run and return the time since its start.
/*!	\return 32bit cycle count
*/
INLINE u32 profile_stop()
{
   REG_TM2CNT= 0;
   return (REG_TM3D<<16)|REG_TM2D;
}

#define SCREEN_WIDTH		256
#define SCREEN_HEIGHT		192
#define WHITE PA_RGB(31, 31, 31)

#define STACK_SIZE		4096

// Structures

// Single co-ordinate
typedef struct {
	s16 x;
	s16 y;
} Point;

// Polygon definition
typedef struct {
	Point origin;
	Point* points;
	u8 pointCount;
} Polygon;

// Floodfill stack
typedef struct {
	u16 items[STACK_SIZE];
	u16 pointer;
} Stack;

// Rotate a polygon
void rotatePolygon(Polygon* source, Polygon* dest, s16 angle) {
	u8 i;
	for (i = 0; i < source->pointCount; i++) {
		dest->points[i].x = (source->points[i].x * (PA_Cos(angle)) - source->points[i].y * (PA_Sin(angle))) >> 8;
		dest->points[i].y = (source->points[i].x * (PA_Sin(angle)) + source->points[i].y * (PA_Cos(angle))) >> 8;
	}
}

// Draw a polygon
void drawPolygon(u8 screen, Polygon* polygon, u16 colour) {
	u8 i;
	s16 x1;
	s16 x2;
	s16 y1;
	s16 y2;

	// Loop through all points drawing from current point to next
	for (i = 0; i < polygon->pointCount; i++) {
		x1 = polygon->points[i].x + polygon->origin.x;
		y1 = polygon->points[i].y + polygon->origin.y;

		// Where is the end point of this line?
		if (i < polygon->pointCount - 1) {
			// Next point
			x2 = polygon->points[i + 1].x + polygon->origin.x;
			y2 = polygon->points[i + 1].y + polygon->origin.y;
		} else {
			// First point
			x2 = polygon->points[0].x + polygon->origin.x;
			y2 = polygon->points[0].y + polygon->origin.y;
		}

		// Draw the line
		PA_Draw16bitLine(screen, x1, y1, x2, y2, colour);
	}
}

// Floodfill Stack functions
u8 popStack(s16* x, s16* y, Stack* stack) { 
	if (stack->pointer > 0) { 
		u16 val = stack->items[stack->pointer]; 
		*y = val >> 8;
		*x = val % SCREEN_WIDTH;
		stack->pointer--;
		return 1;
	} else {
		return 0; 
	}    
}

bool pushStack(s16 x, s16 y, Stack* stack) {
	if (stack->pointer < STACK_SIZE - 1) { 
		stack->pointer++; 
		stack->items[stack->pointer] = x + (y << 8); 
		return 1; 
	} else { 
		return 0; 
	}
}     

void emptyStack(Stack* stack) 
{ 
	s16 x, y; 
	while (popStack(&x, &y, stack)); 
}

void drawLine(u8 screen, s16 x, s16 y, s16 width, u16 colour) {
	u16 drawWidth = (width >> 4);
	u16 row = y * SCREEN_WIDTH;

	u16 i;

	// DMA values
	u16 copyWidth = drawWidth;
	u16* copyFrom = PA_DrawBg[screen] + x + row;
	u16* copyTo = PA_DrawBg[screen] + x + drawWidth + row;

	// Initial row section
	u16* drawStart = PA_DrawBg[screen] + x + row;

	if (drawWidth > 0) {

		// Draw initial pixels by plotting
		for (i = 0; i < drawWidth; i++) {
			*(drawStart + i) = colour;
		}

		// Draw middle sections with DMA
		u8 j;

		for (j = 0; j < 4; j++) {
			DMA_Copy(copyFrom, copyTo, copyWidth, DMA_16NOW);
			copyTo += copyWidth;
			copyWidth += (drawWidth << j);
		}
	}

	// Draw last pixels
	for (i = copyTo - (PA_DrawBg[screen] + x + row); i < width; i++) {
		*(drawStart + i) = colour;
	}
}

// Scanline floodfill algorithm
void floodFill(u8 screen, s16 x, s16 y, u16 newColour, u16 oldColour)
{
	// Exit if colours match
	if (oldColour == newColour) return;

	// Initalise stack
	Stack stack;
	stack.pointer = 0;

	s16 x1; 
	u8 spanUp, spanDown;
	u16 rowThis, rowUp, rowDown;
	s16 rowStart;
	s16 rowWidth;

	// Exit if we can't add to the stack
	if (!pushStack(x, y, &stack)) return;

	// Continue until stack is empty
	while (popStack(&x, &y, &stack)) {

		x1 = x;
		rowThis = y * SCREEN_WIDTH;
		rowUp = (y - 1) * SCREEN_WIDTH;
		rowDown = (y + 1) * SCREEN_WIDTH;

		// Locate leftmost column on screen in this row containing old colour
		while ((x1 >= 0) && (PA_DrawBg[screen][x1 + rowThis] == oldColour)) {
			x1--;
		}

		// Move to first column of old colour
		x1++;

		rowStart = x1;
		rowWidth = 0;

		spanUp = spanDown = 0;

		// Scan right, filling each column of old colour
		while ((x1 < SCREEN_WIDTH) && (PA_DrawBg[screen][x1 + rowThis] == oldColour)) {

			// Check pixel above
			if ((!spanUp) && (y > 0) && (PA_DrawBg[screen][x1 + rowUp] == oldColour)) {
				if (!pushStack(x1, y - 1, &stack)) return;
				spanUp = 1;
			} else if ((spanUp) && (y > 0) && (PA_DrawBg[screen][x1 + rowUp] != oldColour)) {
				spanUp = 0;
			}

			// Check pixel below
			if ((!spanDown) && (y < SCREEN_HEIGHT - 1) && (PA_DrawBg[screen][x1 + rowDown] == oldColour)) {
				if (!pushStack(x1, y + 1, &stack)) return;
				spanDown = 1;
			} else if (spanDown && (y < SCREEN_HEIGHT - 1) && (PA_DrawBg[screen][x1 + rowDown] != oldColour)) {
				spanDown = 0;
			}

			x1++;
			rowWidth++;
		}

		// Draw line
		drawLine(screen, rowStart, y, rowWidth, newColour);
	}
}

// --------------------------------------------------------------------
// Faster replacement 
// --------------------------------------------------------------------

#define REG_DMA3FILL	*(vu32*)0x40000EC

typedef struct POINT8 {	u8 x, y; } POINT8;

#define PUSH_EA(_sp, _x, _y)	do { _sp->x= _x; _sp->y= _y; _sp++; } while(0)
#define POP_EA(_sp, _x, _y)		do { _sp--; _x= _sp->x; _y= _sp->y; } while(0)

// Scanline floodfill algorithm
void floodFillB(int screen, int x, int y, u16 clrNew)
{
	u32 dstW= SCREEN_WIDTH, dstH= SCREEN_HEIGHT;
	u16 *dstD= PA_DrawBg[screen];
	u16 clrOld= dstD[y*dstW+x];
	
	// Exit if colours match
	if(clrOld == clrNew)
		return;

	// Initalize stack
	POINT8 stack[STACK_SIZE], *sp= stack;
	PUSH_EA(sp, x, y);

	// Continue until stack is empty
	while(sp != stack)
	{
		POP_EA(sp, x, y);
		int left= x;
		u16 *dstL= &dstD[y*dstW];

		// Locate leftmost column on screen in this row containing old colour
		while((dstL[left] == clrOld) && (left >= 0))
			left--;
		left++;

		x= left;
		u32 clr, clrUp= -1, clrDown= -1;

		while((dstL[x] == clrOld) && (x < dstW))
		{
			// Check pixel above
			clr= dstL[x-dstW];
			if(clr!=clrUp && (clr == clrOld) && (y > 0))
				PUSH_EA(sp, x, y-1);
			clrUp= clr;

			clr= dstL[x+dstW];
			if(clr!=clrDown && (clr == clrOld) && (y+1<dstH))
				PUSH_EA(sp, x, y+1);
			clrDown= clr;
			
			x++;
		}

		// Draw line
		REG_DMA3FILL= clrNew;
		dmaFillHalfWords(&REG_DMA3FILL, &dstL[left], (x-left)*2); 
	}
}

//! Generalized 16bpp floodfill in asm. Fast and uses real stack.
void bmp16_floodfill_asm(int x, int y, u16 clr, void *dstBase, 
	u16 dstW, u16 dstH);

int main(int argc, char ** argv) {

	PA_Init();		// Initializes PA_Lib
	PA_InitVBL();	// Initializes a standard VBL
	
	PA_Init16bitBg(0, 0);
	PA_Init8bitBg(1, 3);

	// Polygon definitiom
	Polygon rectangle;
	rectangle.origin.x = SCREEN_WIDTH >> 1;
	rectangle.origin.y = SCREEN_HEIGHT >> 1;

	rectangle.pointCount = 4;
	rectangle.points = (Point*)malloc(sizeof(Point) * rectangle.pointCount);
	rectangle.points[0].x = -25;
	rectangle.points[0].y = 25;
	rectangle.points[1].x = 25;
	rectangle.points[1].y = 25;
	rectangle.points[2].x = 25;
	rectangle.points[2].y = -25;
	rectangle.points[3].x = -25;
	rectangle.points[3].y = -25;

	// Store for adjusted polygon
	Polygon worldRect;
	worldRect.origin.x = rectangle.origin.x;
	worldRect.origin.y = rectangle.origin.y;
	worldRect.pointCount = 4;
	worldRect.points = (Point*)malloc(sizeof(Point) * worldRect.pointCount);

	int angle = 2;

	int tm[3];	
	char str[64];
	PA_SetBgPalCol(1, 1, PA_RGB(31, 31, 31));

		DMA_Copy(Blank, PA_DrawBg[0], 49152, DMA_16NOW);

		rotatePolygon(&rectangle, &worldRect, angle);
		drawPolygon(0, &worldRect, WHITE);

		profile_start();
		floodFill(0, worldRect.origin.x, worldRect.origin.y, WHITE, 0);
		tm[0]= profile_stop();
		
		PA_WaitForVBL();
		profile_start();
		floodFillB(0, worldRect.origin.x, worldRect.origin.y, PA_RGB(31,0,0));
		tm[1]= profile_stop();

		PA_WaitForVBL();
		profile_start();
		bmp16_floodfill_asm(worldRect.origin.x, worldRect.origin.y, 
			PA_RGB(0, 31, 0), PA_DrawBg[0], 256, 192);
		tm[2]= profile_stop();

		sprintf(str, "%d / %d / %d", tm[0], tm[1], tm[2]);
		PA_SmartText(1, 0, 64, 255, 129, str, 1, 2, 1, 100);

	while(1);
	
	return 0;
}
